@@ -325,6 +325,14 @@ do_timeout_req_test_for_session (SoupSession *session)
 		errors++;
 		g_clear_error (&error);
 	} else {
+		soup_test_request_read_all (req, stream, NULL, &error);
+		if (error) {
+			debug_printf (1, "  Unexpected error on read: %s\n",
+				      error->message);
+			errors++;
+			g_clear_error (&error);
+		}
+
 		soup_test_request_close_stream (req, stream, NULL, &error);
 		if (error) {
 			debug_printf (1, "  Unexpected error on close: %s\n",
@@ -400,6 +408,7 @@ do_persistent_connection_timeout_test (void)
 	do_timeout_test_for_session (session);
 	soup_test_session_abort_unref (session);
 
+	debug_printf (1, "  Async session, request API\n");
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC,
 					 SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
 					 NULL);
@@ -411,6 +420,7 @@ do_persistent_connection_timeout_test (void)
 	do_timeout_test_for_session (session);
 	soup_test_session_abort_unref (session);
 
+	debug_printf (1, "  Sync session, request API\n");
 	session = soup_test_session_new (SOUP_TYPE_SESSION_SYNC, NULL);
 	do_timeout_req_test_for_session (session);
 	soup_test_session_abort_unref (session);
